// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

import 'package:meta/meta.dart';

import '../event.dart';
import 'room_event.dart';
import '../../identifier.dart';

class MessageEvent extends RoomEvent {
  static const matrixType = 'm.room.message';

  @override
  final String type = matrixType;

  @override
  final MessageEventContent content;

  MessageEvent._(
    RoomEventArgs args, {
    @required this.content,
  }) : super(args);

  factory MessageEvent(
    RoomEventArgs args, {
    @required MessageEventContent content,
  }) {
    if (content is TextMessage) {
      return TextMessageEvent(args, content: content);
    } else if (content is EmoteMessage) {
      return EmoteMessageEvent(args, content: content);
    } else if (content is ImageMessage) {
      return ImageMessageEvent(args, content: content);
    } else if (content is AudioMessage) {
      return AudioMessageEvent(args, content: content);
    } else if (content == null) {
      return MessageEvent._(args, content: null);
    } else {
      // TODO: Raw/custom message events
      return null;
    }
  }
}

abstract class MessageEventContent extends EventContent {
  @protected
  String get type;

  EventId get inReplyToId;

  MessageEventContent();

  factory MessageEventContent.fromJson(Map<String, dynamic> content) {
    if (content == null) return null;

    final msgtype = content['msgtype'];

    EventId inReplyTo;
    if (content.containsKey('m.relates_to') &&
        content['m.relates_to']?.containsKey('m.in_reply_to') == true) {
      final repliesTo = content['m.relates_to']['m.in_reply_to']['event_id'];
      inReplyTo = EventId(repliesTo);
    }

    switch (msgtype) {
      case ImageMessage.matrixMessageType:
        final body = content['body'];

        var url = content['url'];
        if (url == null) {
          return null;
        }

        var info = content['info'];
        if (info != null) {
          var width = info['w'];
          var height = info['h'];

          if (width is! int) {
            width = width?.round();
          }

          if (height is! int) {
            height = height?.round();
          }

          info = ImageInfo(width: width, height: height);
        }

        url = Uri.tryParse(url);

        return ImageMessage(
          body: body,
          url: url,
          info: info,
          inReplyToId: inReplyTo,
        );
      case AudioMessage.matrixMessageType:
        final body = content['body'];

        var url = content['url'];
        if (url == null) {
          return null;
        }

        var info = content['info'];
        if (info != null) {
          info = AudioInfo(
            duration: info['duration'] != null
                ? Duration(milliseconds: info['duration'].round())
                : null,
            mimetype: info['mimetype'],
            size: info['size'],
          );
        }

        url = Uri.tryParse(url);
        return AudioMessage(
          body: body,
          url: url,
          info: info,
          inReplyToId: inReplyTo,
        );
      case TextMessage.matrixMessageType:
      case EmoteMessage.matrixMessageType:
      default:
        final body = content['body'];
        final formattedBody = content['formatted_body'];

        if (msgtype == EmoteMessage.matrixMessageType) {
          return EmoteMessage(
            body: body,
            formattedBody: formattedBody,
            inReplyToId: inReplyTo,
          );
        } else {
          return TextMessage(
            body: body,
            formattedBody: formattedBody,
            inReplyToId: inReplyTo,
          );
        }
    }
  }

  @override
  List<Object> get props => [type, inReplyToId];

  @override
  Map<String, dynamic> toJson() {
    final json = <String, dynamic>{'msgtype': type};

    if (inReplyToId != null) {
      json.addAll({
        'm.relates_to': {
          'm.in_reply_to': {
            'event_id': inReplyToId.toString(),
          }
        }
      });
    }

    return json;
  }
}

class TextMessage extends MessageEventContent {
  static const String matrixMessageType = 'm.text';

  @override
  final String type = matrixMessageType;

  final String body;
  final String formattedBody;

  @override
  final EventId inReplyToId;

  TextMessage({@required this.body, String formattedBody, this.inReplyToId})
      : formattedBody = formattedBody ?? body;

  @override
  List<Object> get props => [...super.props, body, formattedBody];

  @override
  Map<String, dynamic> toJson() =>
      super.toJson()..addAll({'body': body, 'formatted_body': formattedBody});
}

class TextMessageEvent extends MessageEvent {
  @override
  final TextMessage content;

  TextMessageEvent(
    RoomEventArgs args, {
    @required this.content,
  }) : super._(args, content: content);
}

class EmoteMessage extends TextMessage {
  static const String matrixMessageType = 'm.emote';

  @override
  final String type = matrixMessageType;

  EmoteMessage({
    @required String body,
    String formattedBody,
    EventId inReplyToId,
  }) : super(
          body: body,
          formattedBody: formattedBody,
          inReplyToId: inReplyToId,
        );
}

class EmoteMessageEvent extends TextMessageEvent {
  @override
  final EmoteMessage content;

  EmoteMessageEvent(
    RoomEventArgs args, {
    @required this.content,
  }) : super(args, content: content);
}

class ImageMessage extends MessageEventContent {
  static const String matrixMessageType = 'm.image';

  @override
  final String type = matrixMessageType;

  final String body;
  final Uri url;
  final ImageInfo info;

  @override
  final EventId inReplyToId;

  ImageMessage({
    @required this.body,
    this.url,
    this.info,
    this.inReplyToId,
  });

  @override
  List<Object> get props => [...super.props, body, url, info];

  @override
  Map<String, dynamic> toJson() {
    final json = super.toJson()
      ..addAll({
        'body': body,
        'url': url?.toString(),
      });

    if (info != null) {
      final jsonInfo = json['info'] = {};

      if (info.width != null) {
        jsonInfo['w'] = info.width;
      }

      if (info.height != null) {
        jsonInfo['h'] = info.height;
      }
    }

    return json;
  }
}

class ImageInfo {
  final int width;
  final int height;

  ImageInfo({this.width, this.height});
}

class ImageMessageEvent extends MessageEvent {
  @override
  final ImageMessage content;

  ImageMessageEvent(
    RoomEventArgs args, {
    @required this.content,
  }) : super._(args, content: content);
}

class AudioMessage extends MessageEventContent {
  static const String matrixMessageType = 'm.audio';

  @override
  final String type = matrixMessageType;

  final String body;
  final Uri url;
  final AudioInfo info;

  @override
  final EventId inReplyToId;

  AudioMessage({
    @required this.body,
    this.url,
    this.info,
    this.inReplyToId,
  });

  @override
  List<Object> get props => [...super.props, body, url, info];

  @override
  Map<String, dynamic> toJson() {
    final json = <String, dynamic>{
      'body': body,
      'url': url?.toString(),
    };

    if (info != null) {
      json['info'] = {
        'duration': info.duration?.inMilliseconds,
        'mimetype': info.mimetype,
        'size': info.size
      };
    }
    return super.toJson()..addAll(json);
  }
}

class AudioInfo {
  final Duration duration;
  final String mimetype;
  final int size;

  AudioInfo({this.duration, this.mimetype, this.size});
}

class AudioMessageEvent extends MessageEvent {
  @override
  final AudioMessage content;

  AudioMessageEvent(
    RoomEventArgs args, {
    @required this.content,
  }) : super._(args, content: content);
}
