// Copyright (C) 2020  Wilko Manger
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

import 'dart:async';
import 'dart:isolate';

import 'package:meta/meta.dart';

import '../../context.dart';
import '../../event/ephemeral/ephemeral.dart';
import '../../event/event.dart';
import '../../homeserver.dart';
import '../../identifier.dart';
import '../../room/member/member_timeline.dart';
import '../../my_user.dart';
import '../../room/room.dart';
import '../../room/rooms.dart';
import '../../store/store.dart';
import '../../room/timeline.dart';

import '../updater.dart';
import 'instruction.dart';
import 'isolate_runner.dart';

/// Manages updates to [MyUser] in a different [Isolate].
class IsolatedUpdater implements Updater {
  static Future<IsolatedUpdater> create(
    MyUser myUser,
    Homeserver homeserver,
    StoreLocation storeLocation, {
    bool saveMyUserToStore = false,
  }) async {
    final updater = IsolatedUpdater._(
      myUser,
      homeserver,
      storeLocation,
      saveMyUserToStore: saveMyUserToStore,
    );

    await updater._initialized;

    return updater;
  }

  IsolatedUpdater._(
    this._user,
    this.homeserver,
    StoreLocation storeLocation, {
    bool saveMyUserToStore = false,
  }) {
    Updater.register(_user.id, this);

    _messageStream.listen((message) {
      if (message is MinimizedUpdate) {
        final update = message;
        _user = _user.merge(update.delta);

        _subject.add(update.deminimize(_user));
        return;
      }

      if (message is SendPort) {
        _sendPort = message;

        _sendPort.send(
          UpdaterArgs(
            myUser: _user,
            homeserverUrl: homeserver.url,
            storeLocation: storeLocation,
            saveMyUserToStore: saveMyUserToStore,
          ),
        );
      }

      if (message is IsolateInitialized) {
        _initializedCompleter.complete();
      }

      if (message is ErrorWithStackTraceString) {
        _subject.addError(
          message.error,
          StackTrace.fromString(message.stackTraceString),
        );
      }
    });

    Isolate.spawn(IsolateRunner.run, _receivePort.sendPort);
  }

  SendPort _sendPort;

  final _receivePort = ReceivePort();

  Stream<dynamic> __messageStream;
  Stream<dynamic> get _messageStream =>
      __messageStream ??= _receivePort.asBroadcastStream();

  final _initializedCompleter = Completer<void>();
  Future<void> get _initialized => _initializedCompleter.future;

  MyUser _user;
  @override
  MyUser get user => _user;

  @override
  final Homeserver homeserver;

  IsolatedSyncer _syncer;

  @override
  IsolatedSyncer get syncer => _syncer ??= IsolatedSyncer(this);

  final _subject = StreamController<Update>.broadcast();

  @override
  Stream<Update> get updates => _subject.stream;

  /// Sends an instruction to the isolate, possibly with a return value.
  Future<T> _execute<T>(Instruction<T> instruction) async {
    _sendPort.send(instruction);

    if (instruction.expectsReturnValue) {
      final stream =
          instruction is RequestInstruction ? updates : _messageStream;

      return (await stream.firstWhere((event) => event is T)) as T;
    }

    return null;
  }

  Stream<T> _executeStream<T>(
    Instruction<T> instruction, {
    @required updateCount,
  }) {
    _sendPort.send(instruction);

    final stream = instruction is RequestInstruction ? updates : _messageStream;

    return stream
        .where((msg) => msg is T)
        .map((msg) => msg as T)
        .take(updateCount);
  }

  @override
  Future<RequestUpdate<MemberTimeline>> kick(UserId id, {RoomId from}) =>
      _execute(KickInstruction(id, from));

  @override
  Future<RequestUpdate<Timeline>> loadRoomEvents({
    RoomId roomId,
    int count = 20,
  }) =>
      _execute(LoadRoomEventsInstruction(roomId, count));

  @override
  Future<RequestUpdate<MemberTimeline>> loadMembers({
    RoomId roomId,
    int count = 10,
  }) =>
      _execute(LoadMembersInstruction(roomId, count));

  @override
  Future<RequestUpdate<Rooms>> loadRooms(
    Iterable<RoomId> roomIds,
    int timelineLimit,
  ) =>
      _execute(LoadRoomsInstruction(roomIds.toList(), timelineLimit));

  @override
  Future<RequestUpdate<MyUser>> logout() => _execute(LogoutInstruction());

  @override
  Future<RequestUpdate<ReadReceipts>> markRead({
    RoomId roomId,
    EventId until,
    bool receipt = true,
  }) =>
      _execute(MarkReadInstruction(roomId, until, receipt));

  @override
  Stream<RequestUpdate<Timeline>> send(
    RoomId roomId,
    EventContent content, {
    String transactionId,
    String stateKey = '',
  }) =>
      _executeStream(
        SendInstruction(
          roomId,
          content,
          transactionId,
          stateKey,
        ),
        // 2 updates are sent, one for local echo and one for being sent.
        updateCount: 2,
      );

  @override
  Future<RequestUpdate<Ephemeral>> setIsTyping({
    RoomId roomId,
    bool isTyping,
    Duration timeout = const Duration(seconds: 30),
  }) =>
      _execute(SetIsTypingInstruction(roomId, isTyping, timeout));

  @override
  Future<RequestUpdate<MyUser>> setName({String name}) =>
      _execute(SetNameInstruction(name));

  @override
  Future<RequestUpdate<MyUser>> setPusher(Map<String, dynamic> pusher) =>
      _execute(SetPusherInstruction(pusher));
}

class IsolatedSyncer implements Syncer {
  final IsolatedUpdater _updater;

  IsolatedSyncer(this._updater);

  bool _isSyncing = false;

  @override
  bool get isSyncing => _isSyncing;

  @override
  void start({Duration maxRetryAfter = const Duration(seconds: 30)}) {
    _updater._execute(
      StartSyncInstruction(maxRetryAfter),
    );
    _isSyncing = true;
  }

  @override
  Future<void> stop() async {
    await _updater._execute(StopSyncInstruction());
    _isSyncing = false;
  }
}

@immutable
abstract class MinimizedUpdate<T extends Update> {
  final MyUser delta;

  MinimizedUpdate(this.delta);

  T deminimize(MyUser user);
}

class MinimizedSyncUpdate extends MinimizedUpdate<SyncUpdate> {
  MinimizedSyncUpdate({@required MyUser delta}) : super(delta);

  @override
  SyncUpdate deminimize(MyUser user) => SyncUpdate(user, delta);
}

class MinimizedRequestUpdate<T extends Contextual<T>>
    extends MinimizedUpdate<RequestUpdate<T>> {
  final T deltaData;
  final RequestType type;

  MinimizedRequestUpdate({
    @required MyUser delta,
    @required this.deltaData,
    @required this.type,
  }) : super(delta);

  @override
  RequestUpdate<T> deminimize(MyUser user) {
    final deltaData = this.deltaData;

    return RequestUpdate<T>(
      user,
      delta,
      data: deltaData is Contextual<T> ? deltaData.propertyOf(user) : null,
      deltaData: deltaData,
      type: type,
    );
  }
}

class ErrorWithStackTraceString {
  final dynamic error;
  final String stackTraceString;

  ErrorWithStackTraceString(this.error, StackTrace stackTrace)
      : stackTraceString = stackTrace.toString();
}
